"""
SPARTN Lookup and Decode Tables

Created on 10 Feb 2023

Information Sourced from https://www.spartnformat.org/download/
(available in the public domain) © 2021 u-blox AG. All rights reserved.

:author: semuadmin
"""

SF015_ENUM = SF022_ENUM = {
    0: "0 secs",
    1: "1 secs",
    2: "5 secs",
    3: "10 secs",
    4: "30 secs",
    5: "60 secs",
    6: "120 secs",
    7: "320 secs",
}

SF024_ENUM = {
    0: "unknown",
    1: "0.01 m",
    2: "0.02 m",
    3: "0.05 m",
    4: "0.1 m",
    5: "0.3 m",
    6: "1.0 m",
    7: "> 1.0 m",
}

SF042_ENUM = {
    0: "unknown",
    1: "<= 0.010 m",
    2: "<= 0.020 m",
    3: "<= 0.040 m",
    4: "<= 0.080 m",
    5: "<= 0.160 m",
    6: "<= 0.320 m",
    7: "> 0.320 m",
}

SF044_ENUM = {
    0: "Troposphere small coefficient block",
    1: "Troposphere large coefficient block",
}

SF051_ENUM = {
    0: "Troposphere small residual",
    1: "Tropospherelarge residual",
}

SF055_ENUM = {
    0: "Unknown",
    1: "<= 0.03 TECU",
    2: "<= 0.05 TECU",
    3: "<= 0.07 TECU",
    4: "<= 0.14 TECU",
    5: "<= 0.28 TECU",
    6: "<= 0.56 TECU",
    7: "<= 1.12 TECU",
    8: "<= 2.24 TECU",
    9: "<= 4.48 TECU",
    10: "<= 8.96 TECU",
    11: "<= 17.92 TECU",
    12: "<= 35.84 TECU",
    13: "<= 71.68 TECU",
    14: "<= 143.36 TECU",
    15: "> 143.36 TECU",
}

SF056_ENUM = {
    0: "Ionosphere small coefficient block",
    1: "Ionosphere large coefficient block",
}

SF063_ENUM = {
    0: "Ionosphere small residual",
    1: "Ionosphere medium residual",
    2: "Ionosphere large residual",
    3: "Ionosphere extra large residual",
}

SF070_ENUM = {
    0: "350 km",
    1: "400 km",
    2: "450 km",
    3: "500 km",
}

SF077_ENUM = SF078_ENUM = {
    0: "2.5 deg",
    1: "5.0 deg",
    2: "10.0 deg",
    3: "15.0 deg",
}

SF081_ENUM = {
    0: "small VTEC residual",
    1: "large VTEC residual",
}

SF085_ENUM = {
    0: "AES",
    1: "ChaCha12",
    2: "ChaCha20",
    # 3-15 : TBD
}

SF087_ENUM = {  # no bits
    0: 96,
    1: 128,
    2: 192,
    3: 256,
    4: 512,
    # 5-15 : TBD
}

SF090_ENUM = {
    0: "none",
    1: "Ed25519",
    2: "SHA-2",
    3: "SHA-3",
    # 4-15 : TBD
}

SF091_ENUM = {  # no bits
    0: 32,
    1: 64,
    2: 96,
    3: 128,
    4: 192,
    5: 256,
    6: 512,
    # 7-15 : TBD
}

SF093_ENUM = {  # leftmost 2 bits
    0: 36,
    1: 45,
    2: 54,
    3: 64,
}

SF094_ENUM = {  # leftmost 2 bits
    0: 37,
    1: 46,
    2: 55,
    3: 64,
}

SF095_ENUM = {  # leftmost 2 bits
    0: 10,
    1: 40,
    2: 48,
    3: 64,
}

SF096_ENUM = {
    0: "Galileo F/NAV",
    1: "Galileo I/NAV",
    2: "Galileo C/NAV",
    # 3-7: TBD
}

SF097_ENUM = {
    0: "D1 Nav (B1I)",
    1: "D2 Nav (B1I)",
    2: "D1 Nav (B3I)",
    3: "D2 Nav (B3I)",
    4: "B-CNAV1",
    5: "B-CNAV2",
    # 6-15: TBD
}

SF098_ENUM = {
    0: "LNAV (L1C/A)",
    1: "CNAV2 (L1C)",
    2: "CNAV (L2C,L5)",
    # 3-7: TBD
}

SF102_ENUM = {  # rightmost bits left to right
    0: "L1C phase bias",
    1: "L5Q phase bias",
    2: "L7Q phase bias",
    # 3-14: spare phase bias
}

SF103_ENUM = {  # rightmost bits left to right
    0: "L2I phase bias",
    1: "L5P phase bias",
    2: "L7I phase bias",
    3: "L6I phase bias",
    4: "L1P phase bias",
    5: "L7P phase bias",
    6: "L8P phase bias",
    # 7-14: spare phase bias
}

SF104_ENUM = {  # rightmost bits left to right
    0: "L1C phase bias",
    1: "L2L phase bias",
    2: "L5Q phase bias",
    # 3-10: spare phase bias
}

SF105_ENUM = {  # rightmost bits left to right
    0: "C1C code bias",
    1: "C5Q code bias",
    2: "C7Q code bias",
    # 3-14: spare code bias
}

SF106_ENUM = {  # rightmost bits left to right
    0: "C2I code bias",
    1: "C5P code bias",
    2: "C7I code bias",
    3: "C6I code bias",
    4: "C1P code bias",
    5: "C7P code bias",
    6: "C8P code bias",
    # 7-14: spare code bias
}

SF107_ENUM = {  # rightmost bits left to right
    0: "C1C code bias",
    1: "C2L code bias",
    2: "C5Q code bias",
    # 3-10: spare code bias
}



"""
SPARTNMessage class.

The MQTT key, required for payload decryption, can be passed as a keyword
or set up as environment variable MQTTKEY.

Created on 10 Feb 2023

:author: semuadmin
:copyright: SEMU Consulting © 2023
:license: BSD 3-Clause
"""

# pylint: disable=invalid-name too-many-instance-attributes

from datetime import datetime
from os import getenv

from pyspartn.exceptions import (
    ParameterError,
    SPARTNMessageError,
    SPARTNParseError,
    SPARTNTypeError,
)
from pyspartn.spartnhelpers import (
    bitsval,
    convert_timetag,
    decrypt,
    escapeall,
    timetag2date,
    valid_crc,
)
from pyspartn.spartntypes_core import (
    CBBMLEN,
    FL,
    NB,
    PBBMLEN,
    SPARTN_DATA_FIELDS,
    SPARTN_MSGIDS,
    SPARTN_PRE,
    STBMLEN,
    VALCRC,
)
from pyspartn.spartntypes_get import SPARTN_PAYLOADS_GET


class SPARTNMessage:
    """
    SPARTNMessage class.
    """

    def __init__(
        self,
        transport: bytes = None,
        validate: int = VALCRC,
        decode: bool = False,
        key: str = None,
        basedate: object = None,
    ):
        """
        Constructor.

        :param bytes transport: SPARTN message transport (None)
        :param bool validate: validate CRC (True)
        :param bool decode: decrypt and decode payloads (False)
        :param str key: decryption key as hexadecimal string (None)
        :param object basedate: basedate as datetime or 32-bit gnssTimeTag as integer (now)
        :raises: ParameterError if invalid parameters
        """
        # pylint: disable=too-many-arguments

        # object is mutable during initialisation only
        super().__setattr__("_immutable", False)

        self._transport = transport
        if self._transport is None:
            raise SPARTNMessageError("Transport must be provided")

        self._preamble = bitsval(self._transport, 0, 8)
        if self._preamble != SPARTN_PRE:  # not SPARTN
            raise SPARTNParseError(f"Unknown message preamble {self._preamble}")

        self._validate = validate
        self._decode = decode
        self._padding = 0
        basedate = datetime.now() if basedate is None else basedate
        if isinstance(basedate, int):  # 32-bit gnssTimeTag
            self._basedate = timetag2date(basedate)
        else:  # datetime
            self._basedate = basedate
        key = getenv("MQTTKEY", None) if key is None else key
        self._key = None if key is None else bytes.fromhex(key)
        self._iv = None

        if self._decode and self._key is None:
            raise ParameterError("Key must be provided if decoding is enabled")

        self._do_attributes()
        self._immutable = True  # once initialised, object is immutable

    def _do_attributes(self):
        """
        Populate SPARTNMessage attributes from transport.

        :param bytes self._transport: SPARTN message transport
        :raises: SPARTNTypeError
        """

        # start of framestart
        self.msgType = bitsval(self._transport, 8, 7)
        self.nData = bitsval(self._transport, 15, 10)
        self.eaf = bitsval(self._transport, 25, 1)
        self.crcType = bitsval(self._transport, 26, 2)
        self.frameCrc = bitsval(self._transport, 28, 4)

        # start of payDesc
        self.msgSubtype = bitsval(self._transport, 32, 4)
        self.timeTagtype = bitsval(self._transport, 36, 1)
        gln = 32 if self.timeTagtype else 16
        self.gnssTimeTag = bitsval(self._transport, 37, gln)
        pos = 37 + gln
        self.solutionId = bitsval(self._transport, pos, 7)
        self.solutionProcId = bitsval(self._transport, pos + 7, 4)
        pos += 11
        if self.eaf:  # encrypted payload
            self.encryptionId = bitsval(self._transport, pos, 4)
            self.encryptionSeq = bitsval(self._transport, pos + 4, 6)
            self.authInd = bitsval(self._transport, pos + 10, 3)
            self.embAuthLen = bitsval(self._transport, pos + 13, 3)
            pos += 16

        # start of payload
        payload = self._transport[int(pos / 8) : int(pos / 8) + self.nData]

        # start of embAuth
        pos += self.nData * 8
        aln = 0
        if hasattr(self, "authInd"):
            if self.authInd > 1:
                if self.embAuthLen == 0:
                    aln = 64
                elif self.embAuthLen == 1:
                    aln = 94
                elif self.embAuthLen == 2:
                    aln = 128
                elif self.embAuthLen == 3:
                    aln = 256
                elif self.embAuthLen == 4:
                    aln = 512
                self.embAuth = bitsval(self._transport, pos, aln)

        # start of CRC
        pos += aln
        self.crc = bitsval(self._transport, pos, (self.crcType + 1) * 8)

        # validate CRC
        core = self._transport[1 : -(self.crcType + 1)]
        if self._validate & VALCRC:
            if not valid_crc(core, self.crc, self.crcType):
                raise SPARTNMessageError(f"Invalid CRC {self.crc}")

        offset = 0  # payload offset in bits
        index = []  # array of (nested) group indices

        # decrypt payload if encrypted
        if self.eaf and self._decode:
            iv = self._get_iv()
            self._payload = decrypt(payload, self._key, iv)
        else:
            self._payload = payload

        key = ""
        try:
            if self._decode:
                pdict = (
                    self._get_dict()
                )  # get payload definition dict for this message identity
                if pdict is None:  # unknown (or not yet implemented) message identity
                    self._do_unknown()
                    return
                for key in pdict:  # process each attribute in dict
                    (offset, index) = self._set_attribute(offset, pdict, key, index)
                self._padding = self.nData * 8 - offset  # byte alignment padding
        except Exception as err:
            raise SPARTNTypeError(
                (
                    f"Error processing attribute '{key}' "
                    f"in message type {self.identity}"
                )
            ) from err

    def _get_iv(self) -> bytes:
        """
        Create 128-bit Encryption Initialisation Vector.

        :return: IV as bytes
        :rtype: bytes
        """

        if self.timeTagtype:  # 32-bits
            timeTag = self.gnssTimeTag
        else:  # Convert 16-bit timetag to 32 bits
            timeTag = convert_timetag(self.gnssTimeTag, self._basedate)

        iv = (
            (self.msgType << 121)  # TF002 7 bits
            + (self.nData << 111)  # TF003 10 bits
            + (self.msgSubtype << 107)  # TF007 4 bits
            + (timeTag << 75)  # TF009 32 bits
            + (self.solutionId << 68)  # TF010 7 bits
            + (self.solutionProcId << 64)  # TF011 4 bits
            + (self.encryptionId << 60)  # TF012 4 bits
            + (self.encryptionSeq << 54)  # TF012 6 bits
            + 1  # padding to 128 bits
        )
        return iv.to_bytes(16, "big")

    def _set_attribute(self, offset: int, pdict: dict, key: str, index: list) -> tuple:
        """
        Recursive routine to set individual, optional or grouped payload attributes.

        :param int offset: payload offset in bits
        :param dict pdict: dict representing payload definition
        :param str key: attribute keyword
        :param list index: repeating group index array
        :return: (offset, index[])
        :rtype: tuple
        """

        att = pdict[key]  # get attribute type
        if isinstance(att, tuple):  # attribute group
            siz, _ = att
            if isinstance(siz, tuple):  # conditional group of attributes
                (offset, index) = self._set_attribute_optional(att, offset, index)
            else:  # repeating group of attributes
                (offset, index) = self._set_attribute_group(att, offset, index)
        else:  # single attribute
            offset = self._set_attribute_single(att, offset, key, index)

        return (offset, index)

    def _set_attribute_optional(self, attg: tuple, offset: int, index: list) -> tuple:
        """
        Process optional group of attributes, subject to condition being met:
        a) group is present if attribute value = specific value, otherwise absent
        b) group is present if attribute value is in specific range, otherwise absent

        :param tuple attg: attribute group - tuple of ((attribute name, condition), group dict)
        :param int offset: payload offset in bits
        :param list index: repeating group index array
        :return: (offset, index[])
        :rtype: tuple
        """

        pres = False
        (numr, con), gdict = attg  # (attribute, condition), group dictionary
        # "+n" suffix signifies that one or more nested group indices
        # must be appended to name e.g. "DF379_01", "IDF023_03"
        if "+" in numr:
            numr, nestlevel = numr.split("+")
            for i in range(int(nestlevel)):
                numr += f"_{index[i]:02d}"
        if isinstance(con, int):  # present if attribute == value
            pres = getattr(self, numr) == con
        elif isinstance(con, list):  # present if attribute in range of values
            pres = getattr(self, numr) in con

        # recursively process each group attribute,
        # incrementing the payload offset as we go
        if pres:
            for key1 in gdict:
                (offset, index) = self._set_attribute(offset, gdict, key1, index)

        return (offset, index)

    def _set_attribute_group(self, attg: tuple, offset: int, index: list) -> tuple:
        """
        Process (nested) group of attributes. Group size (number of repeats)
        can be signified in a number of ways:
        a) size = fixed integer
        b) size = value of named attribute e.g. SF030
        c) size = number of bits set in named attribute e.g. SF011

        :param tuple attg: attribute group - tuple of (size, group dict)
        :param int offset: payload offset in bits
        :param list index: repeating group index array
        :return: (offset, index[])
        :rtype: tuple
        """

        index.append(0)
        numr, gdict = attg  # size, group dictionary

        # derive or retrieve number of items in group
        if isinstance(numr, int):  # repeats = fixed integer
            rng = numr
        elif isinstance(numr, str):  # repeats defined in named attribute
            # "+n" suffix signifies that one or more nested group indices
            # must be appended to name e.g. "DF379_01", "IDF023_03"
            if "+" in numr:
                numr, nestlevel = numr.split("+")
                for i in range(int(nestlevel)):
                    numr += f"_{index[i]:02d}"
            if numr[0:3] == NB:  # repeats = num bits set
                rng = bin(getattr(self, numr[3:])).count("1")
            else:
                rng = getattr(self, numr)  # repeats = attribute value
                if numr in ("SF030", "SF071"):
                    rng += 1

        # recursively process each group attribute,
        # incrementing the payload offset and index as we go
        for i in range(rng):
            index[-1] = i + 1
            for key1 in gdict:
                (offset, index) = self._set_attribute(offset, gdict, key1, index)

        index.pop()  # remove this (nested) group index

        return (offset, index)

    def _set_attribute_single(
        self,
        att: object,
        offset: int,
        key: str,
        index: list,
    ) -> int:
        """
        Set individual attribute value.

        :param str att: attribute type string e.g. 'INT008'
        :param int offset: payload offset in bits
        :param str key: attribute keyword
        :param list index: repeating group index array
        :return: offset
        :rtype: int
        """
        # pylint: disable=no-member

        # if attribute is part of a (nested) repeating group, suffix name with index
        keyr = key
        for i in index:  # one index for each nested level
            if i > 0:
                keyr += f"_{i:02d}"

        # get value of required number of bits at current payload offset
        # (attribute length, resolution, minimum, description)
        attinfo = SPARTN_DATA_FIELDS[key]
        attlen = attinfo[0]
        atttyp = attinfo[1]  # IN, EN, BM, FL
        if isinstance(attlen, str):  # variable length attribute
            attlen = self._getvarlen(key, index)
        try:
            if atttyp == FL:
                res = attinfo[2]  # resolution (i.e. scaling factor)
                rngmin = attinfo[3]  # range minimum
                val = bitsval(self._payload, offset, attlen, atttyp, res, rngmin)
            else:
                val = bitsval(self._payload, offset, attlen, atttyp)
        except SPARTNMessageError as err:
            # print(self)
            raise err

        setattr(self, keyr, val)

        offset += attlen

        return offset

    def _get_dict(self) -> dict:
        """
        Get payload dictionary corresponding to message identity
        (or None if message type not defined)

        :return: dictionary representing payload definition
        :rtype: dict or None
        """

        pdict = SPARTN_PAYLOADS_GET.get(self.identity, None)
        if pdict == {}:
            pdict = None
        return pdict

    def _getvarlen(self, key: str, index: list) -> int:
        """
        Get length of bitmask attribute based on
        value of preceeding bits.

        :param str keyr: name of attribute
        :param list index: nested group index
        :return: length of attribute in bits
        :rtype: int
        """
        # pylint: disable=no-member, too-many-branches

        # if within repeating group, append nested index
        if len(index) > 0:
            sfx = f"_{index[-1]:02d}"
        else:
            sfx = ""
        attl = 0
        # satellite bitmasks
        if key in ("SF011", "SF012", "SF093", "SF094", "SF095"):
            sflen = getattr(self, STBMLEN + sfx)
            if key == "SF011":  # GPS satellite mask
                attl = [32, 44, 56, 64][sflen]
            elif key == "SF012":  # GLONASS Satellite mask
                attl = [24, 36, 48, 63][sflen]
            elif key == "SF093":  # Galileo satellite mask
                attl = [36, 45, 54, 64][sflen]
            elif key == "SF094":  # BDS satellite mask
                attl = [37, 46, 55, 64][sflen]
            elif key == "SF095":  # QZSS satellite mask
                attl = [10, 40, 48, 64][sflen]
        # phase bias bitmasks
        elif key in ("SF025", "SF026", "SF102", "SF103", "SF104"):
            sflen = getattr(self, PBBMLEN + sfx)
            if key == "SF025":  # GPS phase bias mask
                attl = [6, 11][sflen]
            elif key == "SF026":  # GLONASS phase bias mask
                attl = [5, 9][sflen]
            elif key == "SF102":  # Galileo phase bias mask
                attl = [8, 15][sflen]
            elif key == "SF103":  # BDS phase bias mask
                attl = [8, 15][sflen]
            elif key == "SF104":  # QZSS phase bias mask
                attl = [6, 11][sflen]
        # code bias bitmasks
        elif key in ("SF027", "SF028", "SF105", "SF106", "SF107"):
            sflen = getattr(self, CBBMLEN + sfx)
            if key == "SF027":  # GPS code bias mask
                attl = [6, 11][sflen]
            elif key == "SF028":  # GLONASS code bias mask
                attl = [5, 9][sflen]
            elif key == "SF105":  # Galileo code bias mask
                attl = [8, 15][sflen]
            elif key == "SF106":  # BDS code bias mask
                attl = [8, 15][sflen]
            elif key == "SF107":  # QZSS code bias mask
                attl = [6, 11][sflen]
        elif key == "SF079":  # Grid node present mask
            attl = (getattr(self, f"SF075{sfx}") + 1) * (
                getattr(self, f"SF076{sfx}") + 1
            )
        elif key == "SF088":  # Cryptographic Key length
            attl = [96, 128, 192, 256, 512][self.SF087]
        elif key == "SF092":  # Computed Authentication Data (CAD)
            attl = self.SF091

        return attl

    def _do_unknown(self):
        """
        Handle unknown message type.
        """

    def __str__(self) -> str:
        """
        Human readable representation.

        :return: human readable representation
        :rtype: str
        """

        stg = f"<SPARTN({self.identity}, "
        for i, att in enumerate(self.__dict__):
            if att[0] != "_":  # only show public attributes
                val = self.__dict__[att]
                # escape all byte characters
                if isinstance(val, bytes):
                    val = escapeall(val)
                stg += att + "=" + str(val)
                if i < len(self.__dict__) - 1:
                    stg += ", "
        if self.identity == "UNKNOWN":
            stg += ", Not_Yet_Implemented"
        stg += ")>"

        return stg

    def __repr__(self) -> str:
        """
        Machine readable representation.
        eval(repr(obj)) = obj

        :return: machine readable representation
        :rtype: str
        """

        return f"SPARTNMessage(transport={self._transport})"

    def __setattr__(self, name, value):
        """
        Override setattr to make object immutable after instantiation.

        :param str name: attribute name
        :param object value: attribute value
        :raises: SPARTNMessageError
        """

        if self._immutable:
            raise SPARTNMessageError(
                f"Object is immutable. Updates to {name} not permitted after initialisation."
            )

        super().__setattr__(name, value)

    def serialize(self) -> bytes:
        """
        Serialize message.

        :return: serialized output
        :rtype: bytes
        """

        return self._transport

    @property
    def identity(self) -> str:
        """
        Return message identity.

        :return: message identity e.g. "SPARTN_1X_OCB_GPS"
        :rtype: str
        """

        return SPARTN_MSGIDS.get((self.msgType, self.msgSubtype), "UNKNOWN")

    @property
    def payload(self) -> bytes:
        """
        Return payload.

        :return: payload
        :rtype: bytes
        """

        return self._payload



"""
SPARTN Protocol core globals and constants

Created on 10 Feb 2023

Information Sourced from https://www.spartnformat.org/download/
(available in the public domain) © 2021 u-blox AG. All rights reserved.

:author: semuadmin
"""

# pylint: disable=line-too-long

from datetime import datetime, timezone

TIMEBASE = datetime(2010, 1, 1, 0, 0, tzinfo=timezone.utc)
ERRRAISE = 2
ERRLOG = 1
ERRIGNORE = 0
VALNONE = 0
VALCRC = 1
VALMSGID = 2
SPARTN_PRE = 0x73
SPARTN_PREB = b"s"

# Attribute types
IN = "IN"  # integer
EN = "EN"  # enumeration
BM = "BM"  # bitmask
FL = "FL"  # float

# Transient attribute names used to store variable bitmask length flags
NB = "NB_"
STBMLEN = "SatBitmaskLen"
PBBMLEN = "PhaseBiasBitmaskLen"
CBBMLEN = "CodeBiasBitmaskLen"

# UTC + leap second basedate shift for different constellations
# relative to GPS (UTC + 18 leap seconds); valid as from 2017/1/1
TIMETAGSHIFT = {
    0: 0,  # GPS
    1: 10782,  # GLO = GPS + 3600 * 3 - 18
    2: 0,  # GAL = GPS
    3: -14,  # BEI = GPS - 14
    4: 0,  # QZS = GPS
}

# SPARTN message types
SPARTN_MSGIDS = {
    0: "SPARTN-1X-OCB",  # Orbit, Clock, Bias
    (0, 0): "SPARTN-1X-OCB-GPS",
    (0, 1): "SPARTN-1X-OCB-GLO",
    (0, 2): "SPARTN-1X-OCB-GAL",
    (0, 3): "SPARTN-1X-OCB-BEI",
    (0, 4): "SPARTN-1X-OCB-QZS",
    1: "SPARTN-1X-HPAC",  # High-precision atmosphere correction
    (1, 0): "SPARTN-1X-HPAC-GPS",
    (1, 1): "SPARTN-1X-HPAC-GLO",
    (1, 2): "SPARTN-1X-HPAC-GAL",
    (1, 3): "SPARTN-1X-HPAC-BEI",
    (1, 4): "SPARTN-1X-HPAC-QZS",
    2: "SPARTN-1X-GAD",  # Geographic Area Definition
    (2, 0): "SPARTN-1X-GAD",
    3: "SPARTN-1X-BPAC",  # Basic-precision atmosphere correction
    (3, 0): "SPARTN-1X-BPAC",
    4: "SPARTN-1X-EAS",  # Encryption and Authentication Support
    (4, 0): "SPARTN-1X-EAS-DYN",
    (4, 1): "SPARTN-1X-EAS-GRP",  # deprecated
    120: "SPARTN_1X-PROP",  # Proprietary messages
    (120, 0): "SPARTN-1X-PROP-TEST",
    (120, 1): "SPARTN-1X-PROP-UBLOX",
    (120, 2): "SPARTN-1X-PROP-SWIFT",
}

# Datafields used in message definitions
# key (IN, BM, EN): (length in bits, type, resolution or n/a, description)
# key (FL): (length in bits, type, resolution, range minimum, description)
SPARTN_DATA_FIELDS = {
    "SF005": (9, IN, 1, "Solution issue of update (SIOU)"),
    "SF008": (1, EN, "n/a", "Yaw present flag"),
    "SF009": (1, IN, 1, "Satellite reference datum"),
    "SF010": (1, IN, "n/a", "End of OCB set (EOS)"),
    STBMLEN: (2, IN, "n/a", "Length of satellite bitmask"),
    "SF011": ("34 to 66", BM, "Bitmask", "GPS Satellite mask"),
    "SF012": ("26 to 65", BM, "Bitmask", "GLONASS Satellite mask"),
    "SF013": (1, IN, "n/a", "Do not use (DNU)"),
    "SF014O": (1, BM, "Bitmask", "Orbit data present flag"),
    "SF014C": (1, BM, "Bitmask", "Clock data present flag"),
    "SF014B": (1, BM, "Bitmask", "Bias data present flag"),
    "SF015": (3, IN, "n/a", "Continuity indicator"),
    "SF016": (2, EN, "n/a", "GPS ephemeris type"),
    "SF017": (2, EN, "n/a", "GLO ephemeris type"),
    "SF018": (8, IN, 1, "GPS IODE"),
    "SF019": (7, IN, 1, "GLO IODE"),
    "SF020R": (14, FL, 0.002, -16.382, "Satellite radial corrections"),
    "SF020A": (14, FL, 0.002, -16.382, "Satellite along-track corrections"),
    "SF020C": (14, FL, 0.002, -16.382, "Satellite cross-track corrections"),
    "SF020CK": (14, FL, 0.002, -16.382, "Satellite clock corrections"),
    "SF020PB": (14, FL, 0.002, -16.382, "Phase bias correction"),
    "SF021": (6, FL, 6, 0, "Satellite yaw"),
    "SF022": (3, IN, "n/a", "IODE continuity"),
    "SF023": (1, EN, "n/a", "Fix flag"),
    "SF024": (3, IN, "n/a", "User range error (URE)"),
    PBBMLEN: (1, IN, "n/a", "Length of phase bias bitmask"),
    "SF025": ("7 or 12", BM, "bitmask", "GPS phase bias mask"),
    "SF026": ("6 or 10", BM, "bitmask", "GLONASS phase bias mask"),
    CBBMLEN: (1, IN, "n/a", "Length of code bias bitmask"),
    "SF027": ("7 or 12", BM, "bitmask", "GPS code bias mask"),
    "SF028": ("6 or 10", BM, "bitmask", "GLONASS code bias mask"),
    "SF029": (11, FL, 0.02, -20.46, "Code bias correction"),
    "SF030": (5, IN, 1, "Area Count"),  # NB: area count = SF030 + 1
    "SF031": (8, IN, 1, "Area ID"),
    "SF032": (11, FL, 0.1, -90, "Area reference latitude"),
    "SF033": (12, FL, 0.1, -180, "Area reference longitude"),
    "SF034": (3, IN, 1, "Area latitude grid node count"),
    "SF035": (3, IN, 1, "Area longitude grid node count"),
    "SF036": (5, FL, 0.1, 0.1, "Area latitude grid node spacing"),
    "SF037": (5, FL, 0.1, 0.1, "Area longitude grid node spacing"),
    "SF039": (7, IN, 1, "Number of grid points present"),
    "SF040": (2, IN, 1, "Poly/Grid block present indicator"),
    "SF040T": (2, IN, 1, "Troposphere Poly/Grid block present indicator"),
    "SF040I": (2, IN, 1, "Ionoshere Poly/Grid block present indicator"),
    "SF041": (3, EN, 1, "Troposphere equation type"),
    "SF042": (3, IN, 1, "Troposphere quality"),
    "SF043": (8, FL, 0.004, -0.508, "Area average vertical hydrostatic delay"),
    "SF044": (1, IN, 1, "Troposphere polynomial coefficient size indicator"),
    "SF045": (7, FL, 0.004, -0.252, "Small troposphere coefficient T00"),
    "SF046a": (7, FL, 0.001, -0.063, "Small troposphere coefficient T01"),
    "SF046b": (7, FL, 0.001, -0.063, "Small troposphere coefficient T10"),
    "SF047": (9, FL, 0.0002, -0.0510, "Small troposphere coefficient T11"),
    "SF048": (9, FL, 0.004, -1.020, "Large troposphere coefficient T00"),
    "SF049a": (9, FL, 0.001, -0.255, "Large troposphere coefficient T01"),
    "SF049b": (9, FL, 0.001, -0.255, "Large troposphere coefficient T10"),
    "SF050": (11, FL, 0.0002, -0.2046, "Large troposphere coefficient T11"),
    "SF051": (1, IN, 1, "Troposphere residual field size"),
    "SF052": (6, FL, 0.004, -0.124, "Small troposphere residual zenith delay"),
    "SF053": (8, FL, 0.004, -0.508, "Large troposphere residual zenith delay"),
    "SF054": (3, EN, 1, "Ionosphere equation type"),
    "SF055": (4, IN, 1, "Ionosphere quality"),
    "SF056": (1, IN, 1, "Ionosphere polynomial coefficient size indicator"),
    "SF057": (12, FL, 0.04, -81.88, "Small ionosphere coefficient C00"),
    "SF058a": (12, FL, 0.008, -16.376, "Small ionosphere coefficient C01"),
    "SF058b": (12, FL, 0.008, -16.376, "Small ionosphere coefficient C10"),
    "SF059": (13, FL, 0.002, -8.190, "Small ionosphere coefficient C11"),
    "SF060": (14, FL, 0.04, -327.64, "Large ionosphere coefficient C00"),
    "SF061a": (14, FL, 0.008, -65.528, "Large ionosphere coefficient C01"),
    "SF061b": (14, FL, 0.008, -65.528, "Large ionosphere coefficient C10"),
    "SF062": (15, FL, 0.002, -32.766, "Large ionosphere coefficient C11"),
    "SF063": (2, IN, 1, "Ionosphere residual field size"),
    "SF064": (4, FL, 0.04, -0.28, "Small ionosphere residual slant delay"),
    "SF065": (7, FL, 0.04, -2.52, "Medium ionosphere residual slant delay"),
    "SF066": (10, FL, 0.04, -20.44, "Large ionosphere residual slant delay"),
    "SF067": (14, FL, 0.04, -327.64, "Extra-large ionosphere residual slant delay"),
    "SF068": (4, IN, 1, "Area Issue of Update (AIOU)"),
    "SF069": (1, IN, "n/a", "Reserved"),
    "SF070": (2, IN, 1, "Ionosphere shell height"),
    "SF071": (2, IN, 1, "BPAC area count"),
    "SF072": (2, IN, 1, "BPAC area ID"),
    "SF073": (8, FL, 1.0, -85.0, "BPAC area reference latitude"),
    "SF074": (9, FL, 1.0, -180.0, "BPAC area reference longitude"),
    "SF075": (4, IN, 1, "BPAC area latitude grid node count"),
    "SF076": (4, IN, 1, "BPAC area longitude grid node count"),
    "SF077": (2, IN, 1, "BPAC area latitude grid node spacing "),
    "SF078": (2, IN, 1, "BPAC area longitude grid node spacing "),
    "SF079": ("N", BM, "Bitmask", "Grid node present mask"),
    "SF080": (12, FL, 0.25, -511.75, "Area average VTEC"),
    "SF081": (1, IN, 1, "VTEC size indicator"),
    "SF082": (7, FL, 0.25, -15.75, "Small VTEC residual "),
    "SF083": (11, FL, 0.25, -255.75, "Large VTEC residual "),
    "SF084": (20, IN, 1, "Customer Key ID"),
    "SF085": (4, EN, 1, "Encryption Type"),
    "SF086": (6, IN, 1, "Week of Applicability"),
    "SF087": (4, IN, 1, "Key length"),
    "SF088": ("Key length (SF087)", IN, "1", "Cryptographic Key"),
    "SF089": (5, IN, 1, "Count of Message IDs"),
    "SF090": (4, EN, 1, "Group Authentication Type"),
    "SF091": (4, IN, 1, "Computed Authentication Data (CAD) Length"),
    "SF092": ("CAD length (SF091)", IN, 1, "Computed Authentication Data (CAD)"),
    "SF093": ("38 to 66", BM, "Bitmask", "Galileo satellite mask"),
    "SF094": ("39 to 66", BM, "Bitmask", "BDS satellite mask"),
    "SF095": ("12 to 66", BM, "Bitmask", "QZSS satellite mask"),
    "SF096": (3, EN, "n/a", "Galileo ephemeris type"),
    "SF097": (4, EN, "n/a", "BDS ephemeris type"),
    "SF098": (3, EN, "n/a", "QZSS ephemeris type"),
    "SF099": (10, IN, 1, "Galileo IODnav"),
    "SF100": (8, IN, 1, "BDS IODE/IODC"),
    "SF101": (8, IN, 1, "QZSS IODE"),
    "SF102": ("9 or 16", BM, "Bitmask", "Galileo phase bias mask"),
    "SF103": ("9 or 16", BM, "Bitmask", "BDS phase bias mask"),
    "SF104": ("7 or 12", BM, "Bitmask", "QZSS phase bias mask"),
    "SF105": ("9 or 16", BM, "Bitmask", "Galileo code bias mask"),
    "SF106": ("9 or 16", BM, "Bitmask", "BDS code bias mask"),
    "SF107": ("7 or 12", BM, "Bitmask", "QZSS code bias mask"),
}



"""
SPARTN Protocol core globals and constants

Created on 10 Feb 2023

Information Sourced from https://www.spartnformat.org/download/
(available in the public domain) © 2021 u-blox AG. All rights reserved.

:author: semuadmin
"""

# pylint: disable=too-many-lines, line-too-long

from pyspartn.spartntypes_core import CBBMLEN, NB, PBBMLEN, STBMLEN

OCB_HDR = {  # OCB Header
    "SF005": "Solution issue of update (SIOU)",
    "SF010": "End of set",
    "SF069": "Reserved",
    "SF008": "Yaw present flag",
    "SF009": "Satellite reference datum",
}

HPAC_HDR = {  # HPAC Header
    "SF005": "Solution issue of update (SIOU)",
    "SF068": "Area issue of update (AIOU)",
    "SF069": "Reserved",
    "SF030": "Area count",  # NB: N - 1
}

OCB_SAT_FLAGS = {  # table 6.4
    "SF014O": "Orbit data present flag",
    "SF014C": "Clock data present flag",
    "SF014B": "Bias data present flag",
    "SF015": "Continuity indicator",
}

OCB_ORBIT_BLOCK = {  # part of table 6.5 Orbit Block
    "SF020R": "Orbit radial correction",
    "SF020A": "Orbit along-track correction",
    "SF020C": "Orbit cross-track correction",
    "optSF008-1": (
        ("SF008", 1),  # if SF008 = 1
        {
            "SF021": "Satellite yaw",
        },
    ),
}

OCB_CLOCK_BLOCK = {  # table 6.6 Clock Block
    "optClock": (
        ("SF014C+1", 1),  # if SF014C = 1
        {
            "SF022": "IODE continuity",
            "SF020CK": "Clock correction",
            "SF024": "User range error",
        },
    ),
}

PHAS_BIAS_BLOCK = {  # table 6.12 Phase Bias Block
    "SF023": "Fix flag",
    "SF015": "Continuity indicator",
    "SF020PB": "Phase bias correction",
}

AREA_DATA_BLOCK = {  # table 6.15 Area Data Block
    "SF031": "Area ID",
    "SF039": "Number of grid points present",
    "SF040T": "Tropo blocks indicator",
    "SF040I": "Iono blocks indicator",
}

TROP_DATA_BLOCK = {  # table 6.16 Troposphere Data Block
    "optSF040T-12": (
        ("SF040T+1", [1, 2]),  # if SF040T in 1,2
        {
            "SF041": "Troposhere equation type",
            "SF042": "Troposphere quality",
            "SF043": "Area average vertical hydrostatic delay",
            "SF044": "Troposphere polynomial coefficient size indicator",
            "optSF044-0": (
                ("SF044+1", 0),  # if SF044 = 0 table 6.17
                {
                    "SF045": "Troposphere coefficient T00",
                    "optSF041-12": (
                        ("SF041+1", [1, 2]),  # if SF041 in 1,2
                        {
                            "SF046a": "Troposphere coefficient T01",
                            "SF046b": "Troposphere coefficient T10",
                        },
                    ),
                    "optSF041-2": (
                        ("SF041+1", 2),  # if SF041 = 2
                        {
                            "SF047": "Troposphere coefficient T11",
                        },
                    ),
                },
            ),
            "optSF044-1": (
                ("SF044+1", 1),  # if SF044 = 1 table 6.18
                {
                    "SF048": "Troposphere coefficient T00",
                    "optSF041-12": (
                        ("SF041+1", [1, 2]),  # if SF041 in 1,2
                        {
                            "SF049a": "Troposphere coefficient T01",
                            "SF049b": "Troposphere coefficient T10",
                        },
                    ),
                    "optSF041-2": (
                        ("SF041+1", 2),  # if SF041 = 2
                        {
                            "SF050": "Troposphere coefficient T11",
                        },
                    ),
                },
            ),
        },
    ),
    "optSF040T-2": (  # if SF040T = 2
        ("SF040T+1", 2),  # if SF040T = 2
        {
            "SF051": "Troposphere residual field size",
            "optSF051-0": (
                ("SF051", 0),  # if SF051 = 0
                {"SF052": "Troposphere grid residuals"},
            ),
            "optSF051-1": (  # if SF051 = 1
                ("SF051", 1),  # if SF051 = 1
                {"SF053": "Troposphere grid residuals"},
            ),
        },
    ),
}

ION_SAT_BLOCK = {  # table 6.20 Ionosphere Satellite Block
    "optSF041-12": (
        ("SF041+1", [1, 2]),  # if SF041I in 1,2
        {
            "SF055": "Ionosphere quality",
            "SF056": "Ionosphere polynomial coefficient size indicator",
            "optSF064-0": (
                ("SF056+2", 0),  # if SF056 = 0 table 6.21
                {
                    "SF057": "Ionosphere coefficient C00",
                    "optSF054-12": (
                        ("SF054+1", [1, 2]),  # if SF054 in 1,2
                        {
                            "SF058a": "Ionosphere coefficient C01",
                            "SF058b": "Ionosphere coefficient C10",
                        },
                    ),
                    "optSF054-2": (
                        ("SF054+1", 2),  # if SF054 = 2
                        {
                            "SF059": "Ionosphere coefficient C11",
                        },
                    ),
                },
            ),
            "optSF056-1": (
                ("SF056+2", 1),  # if SF056 = 1 table 6.22
                {
                    "SF060": "Ionosphere coefficient C00",
                    "optSF054-12": (
                        ("SF054+1", [1, 2]),  # if SF054 in 1,2
                        {
                            "SF061a": "Ionosphere coefficient C01",
                            "SF061b": "Ionosphere coefficient C10",
                        },
                    ),
                    "optSF054-2": (
                        ("SF054+1", 2),  # if SF054 = 2
                        {
                            "SF062": "Ionosphere coefficient C11",
                        },
                    ),
                },
            ),
        },
    ),
    "optSF041-2": (
        ("SF041+1", 2),  # if SF041I = 2
        {
            "SF063": "Ionosphere residual field size",
            "optSF063-0": (
                ("SF063", 0),  # if SF063 = 0
                {"SF064": "Ionosphere grid residuals"},
            ),
            "optSF063-1": (
                ("SF063", 1),  # if SF063 = 1
                {"SF065": "Ionosphere grid residuals"},
            ),
            "optSF063-2": (
                ("SF063", 2),  # if SF063 = 2
                {"SF066": "Ionosphere grid residuals"},
            ),
            "optSF063-3": (
                ("SF063", 3),  # if SF063 = 3
                {"SF067": "Ionosphere grid residuals"},
            ),
        },
    ),
}

# ************************************************************************
# SPARTN MESSAGE PAYLOAD DEFINITIONS (NB: PAYLOAD MUST BE DECRYPTED FIRST)
# ************************************************************************

SPARTN_PAYLOADS_GET = {
    # ********************************************************************
    # OCB
    # ********************************************************************
    "SPARTN-1X-OCB-GPS": {
        **OCB_HDR,
        "SF016": "GPS Ephemeris type",
        STBMLEN: "GPS Satellite mask length",
        "SF011": "GPS Satellite mask",
        "groupSat": (  # Satellite Block table 6.4
            NB + "SF011",  # repeating group * num bits set in SF011
            {
                "SF013": "Do not use (DNU)",
                "optSat": (
                    ("SF013+1", 0),  # if SF013 = 0
                    {
                        **OCB_SAT_FLAGS,
                        "optOrbit": (  # # table 6.5 Orbit Block
                            ("SF014O+1", 1),  # if SF014O = 1
                            {
                                "SF018": "GPS IODE",
                                **OCB_ORBIT_BLOCK,
                            },
                        ),
                        **OCB_CLOCK_BLOCK,
                        "optBias": (
                            ("SF014B+1", 1),  # if SF014B = 1
                            {
                                PBBMLEN: "Phase bias mask length",
                                "SF025": "GPS phase bias mask",
                                "groupSF025-BITS": (
                                    NB
                                    + "SF025+1",  # repeating group * num bits set in SF025
                                    {
                                        **PHAS_BIAS_BLOCK,
                                    },
                                ),
                                CBBMLEN: "Code bias mask length",
                                "SF027": "GPS code bias mask",
                                "groupSF027-BITS": (
                                    NB
                                    + "SF027+1",  # repeating group * num bits set in SF027
                                    {
                                        "SF029": "Code bias correction",
                                    },
                                ),
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-OCB-GLO": {
        **OCB_HDR,
        "SF017": "GLO Ephemeris type",
        STBMLEN: "GPS Satellite mask length",
        "SF012": "GLO Satellite mask",
        "groupSat": (  # Satellite Block table 6.4
            NB + "SF012",  # repeating group * num bits set in SF012
            {
                "SF013": "Do not use (DNU)",
                "optSat": (
                    ("SF013+1", 0),  # if SF013 = 0
                    {
                        **OCB_SAT_FLAGS,
                        "optOrbit": (  # table 6.5 Orbit Block
                            ("SF014O+1", 1),  # if SF014O = 1
                            {
                                "SF019": "GLONASS IODE",
                                **OCB_ORBIT_BLOCK,
                            },
                        ),
                        **OCB_CLOCK_BLOCK,
                        "optBias": (
                            ("SF014B+1", 1),  # if SF014B = 1
                            {
                                PBBMLEN: "Phase bias mask length",
                                "SF026": "GLONASS phase bias mask",
                                "groupSF026-BITS": (
                                    NB
                                    + "SF026+1",  # repeating group * num bits set in SF026
                                    {
                                        **PHAS_BIAS_BLOCK,
                                    },
                                ),
                                CBBMLEN: "Code bias mask length",
                                "SF028": "GLONASScode bias mask",
                                "groupSF028-BITS": (
                                    NB
                                    + "SF028+1",  # repeating group * num bits set in SF028
                                    {
                                        "SF029": "Code bias correction",
                                    },
                                ),
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-OCB-GAL": {
        **OCB_HDR,
        "SF096": "GALILEO Ephemeris type",
        STBMLEN: "GPS Satellite mask length",
        "SF093": "GALILEO Satellite mask",
        "groupSat": (  # Satellite Block table 6.4
            NB + "SF093",  # repeating group * num bits set in SF093
            {
                "SF013": "Do not use (DNU)",
                "optSat": (
                    ("SF013+1", 0),  # if SF013 = 0
                    {
                        **OCB_SAT_FLAGS,
                        "optOrbit": (  # table 6.5 Orbit Block
                            ("SF014O+1", 1),  # if SF014O = 1
                            {
                                "SF099": "GALILEO IODE",
                                **OCB_ORBIT_BLOCK,
                            },
                        ),
                        **OCB_CLOCK_BLOCK,
                        "optBias": (
                            ("SF014B+1", 1),  # if SF014B = 1
                            {
                                PBBMLEN: "Phase bias mask length",
                                "SF102": "GALILEO phase bias mask",
                                "groupSF102-BITS": (
                                    NB
                                    + "SF102+1",  # repeating group * num bits set in SF0102
                                    {
                                        **PHAS_BIAS_BLOCK,
                                    },
                                ),
                                CBBMLEN: "Code bias mask length",
                                "SF105": "GALILEO code bias mask",
                                "groupSF105-BITS": (
                                    NB
                                    + "SF105+1",  # repeating group * num bits set in SF0105
                                    {
                                        "SF029": "Code bias correction",
                                    },
                                ),
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-OCB-BEI": {
        **OCB_HDR,
        "SF097": "BEIDOU Ephemeris type",
        STBMLEN: "GPS Satellite mask length",
        "SF094": "BEIDOU Satellite mask",
        "groupSat": (  # Satellite Block table 6.4
            NB + "SF094",  # repeating group * num bits set in SF094
            {
                "SF013": "Do not use (DNU)",
                "optSat": (
                    ("SF013+1", 0),  # if SF013 = 0
                    {
                        **OCB_SAT_FLAGS,
                        "optOrbit": (  # table 6.5 Orbit Block
                            ("SF014O+1", 1),  # if SF014O = 1
                            {
                                "SF100": "BEIDOU IODE",
                                **OCB_ORBIT_BLOCK,
                            },
                        ),
                        **OCB_CLOCK_BLOCK,
                        "optBias": (
                            ("SF014B+1", 1),  # if SF014B = 1
                            {
                                PBBMLEN: "Phase bias mask length",
                                "SF103": "BEIDOU phase bias mask",
                                "groupSF103-BITS": (
                                    NB
                                    + "SF103+1",  # repeating group * num bits set in SF0103
                                    {
                                        **PHAS_BIAS_BLOCK,
                                    },
                                ),
                                CBBMLEN: "Code bias mask length",
                                "SF106": "BEIDOU code bias mask",
                                "groupSF106-BITS": (
                                    NB
                                    + "SF106+1",  # repeating group * num bits set in SF0106
                                    {
                                        "SF029": "Code bias correction",
                                    },
                                ),
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-OCB-QZS": {
        **OCB_HDR,
        "SF098": "QZSS Ephemeris type",
        STBMLEN: "GPS Satellite mask length",
        "SF095": "QZSS Satellite mask",
        "groupSat": (  # Satellite Block table 6.4
            NB + "SF095",  # repeating group * num bits set in SF095
            {
                "SF013": "Do not use (DNU)",
                "optSat": (
                    ("SF013+1", 0),  # if SF013 = 0
                    {
                        **OCB_SAT_FLAGS,
                        "optOrbit": (  # table 6.5 Orbit Block
                            ("SF014O+1", 1),  # if SF014O = 1
                            {
                                "SF101": "QZSS IODE",
                                **OCB_ORBIT_BLOCK,
                            },
                        ),
                        **OCB_CLOCK_BLOCK,
                        "optBias": (
                            ("SF014B+1", 1),  # if SF014B = 1
                            {
                                PBBMLEN: "Phase bias mask length",
                                "SF104": "QZSS phase bias mask",
                                "groupSF104-BITS": (
                                    NB
                                    + "SF104+1",  # repeating group * num bits set in SF0104
                                    {
                                        **PHAS_BIAS_BLOCK,
                                    },
                                ),
                                CBBMLEN: "Code bias mask length",
                                "SF107": "QZSS code bias mask",
                                "groupSF107-BITS": (
                                    NB
                                    + "SF107+1",  # repeating group * num bits set in SF0107
                                    {
                                        "SF029": "Code bias correction",
                                    },
                                ),
                            },
                        ),
                    },
                ),
            },
        ),
    },
    # ********************************************************************
    # HPAC
    # ********************************************************************
    "SPARTN-1X-HPAC-GPS": {
        **HPAC_HDR,
        "groupAtm": (  # Atmosphere Data Block repeating group * (SF030 + 1)
            "SF030",
            {
                **AREA_DATA_BLOCK,
                **TROP_DATA_BLOCK,
                "optSF040I-12": (  # table 6.19 Ionosphere Data Block
                    ("SF040I+1", [1, 2]),  # if SF040I in 1,2
                    {
                        "SF054": "Ionosphere equation type",
                        STBMLEN: "GPS Satellite mask length",
                        "SF011": "GPS Satellite mask",
                        "groupSF011": (  # repeating group * num bits set in SF011
                            NB + "SF011+1",
                            {
                                **ION_SAT_BLOCK,
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-HPAC-GLO": {
        **HPAC_HDR,
        "groupAtm": (  # Atmosphere Data Block repeating group * (SF030 + 1)
            "SF030",
            {
                **AREA_DATA_BLOCK,
                **TROP_DATA_BLOCK,
                "optSF040I-12": (  # table 6.19 Ionosphere Data Block
                    ("SF040I+1", [1, 2]),  # if SF040I in 1,2
                    {
                        "SF054": "Ionosphere equation type",
                        STBMLEN: "GPS Satellite mask length",
                        "SF012": "GLONASS Satellite mask",
                        "groupSF012": (  # repeating group * num bits set in SF012
                            NB + "SF012+1",
                            {
                                **ION_SAT_BLOCK,
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-HPAC-GAL": {
        **HPAC_HDR,
        "groupAtm": (  # Atmosphere Data Block repeating group * (SF030 + 1)
            "SF030",
            {
                **AREA_DATA_BLOCK,
                **TROP_DATA_BLOCK,
                "optSF040I-12": (  # table 6.19 Ionosphere Data Block
                    ("SF040I+1", [1, 2]),  # if SF040I in 1,2
                    {
                        "SF054": "Ionosphere equation type",
                        STBMLEN: "GPS Satellite mask length",
                        "SF093": "GALILEO Satellite mask",
                        "groupSF093": (  # repeating group * num bits set in SF093
                            NB + "SF093+1",
                            {
                                **ION_SAT_BLOCK,
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-HPAC-BEI": {
        **HPAC_HDR,
        "groupAtm": (  # Atmosphere Data Block repeating group * (SF030 + 1)
            "SF030",
            {
                **AREA_DATA_BLOCK,
                **TROP_DATA_BLOCK,
                "optSF040I-12": (  # table 6.19 Ionosphere Data Block
                    ("SF040I+1", [1, 2]),  # if SF040I in 1,2
                    {
                        "SF054": "Ionosphere equation type",
                        STBMLEN: "GPS Satellite mask length",
                        "SF094": "BEIDOU Satellite mask",
                        "groupSF094": (  # repeating group * num bits set in SF094
                            NB + "SF094+1",
                            {
                                **ION_SAT_BLOCK,
                            },
                        ),
                    },
                ),
            },
        ),
    },
    "SPARTN-1X-HPAC-QZS": {
        **HPAC_HDR,
        "groupAtm": (  # Atmosphere Data Block repeating group * (SF030 + 1)
            "SF030",
            {
                **AREA_DATA_BLOCK,
                **TROP_DATA_BLOCK,
                "optSF040I-12": (  # table 6.19 Ionosphere Data Block
                    ("SF040I+1", [1, 2]),  # if SF040I in 1,2
                    {
                        "SF054": "Ionosphere equation type",
                        STBMLEN: "GPS Satellite mask length",
                        "SF095": "QZSS Satellite mask",
                        "groupSF095": (  # repeating group * num bits set in SF095
                            NB + "SF095+1",
                            {
                                **ION_SAT_BLOCK,
                            },
                        ),
                    },
                ),
            },
        ),
    },
    # ********************************************************************
    # GAD
    # ********************************************************************
    "SPARTN-1X-GAD": {
        "SF005": "Solution issue of updated (SIOU)",
        "SF068": "Area issue of update (AIOU)",
        "SF069": "Reserved",
        "SF030": "Area Count",  # NB:  N - 1
        "group": (  # repeating group * (SF030 + 1)
            "SF030",
            {
                "SF031": "Area ID",
                "SF032": "Area reference latitude",
                "SF033": "Area reference longitude",
                "SF034": "Are latitude grid node count",
                "SF035": "Area longitude grid node count",
                "SF036": "Area latitude grid node spacing",
                "SF037": "Area longitude grid node spacing",
            },
        ),
    },
    # ********************************************************************
    # BPAC TODO not yet tested - no available test data source
    # ********************************************************************
    "SPARTN-1X-BPAC": {
        "SF005": "Solution issue of updated (SIOU)",
        "SF069": "Reserved",
        "SF070": "Ionosphere shell height",
        "SF071": "BPAC area count",  # NB: N - 1
        "groupBPAC": (  # repeating group * (SF071 + 1)
            "SF071",
            {
                "SF072": "BPAC area ID",
                "SF073": "BPAC area reference latitude",
                "SF074": "BPAC area reference longitude",
                "SF075": "BPAC area latitude grid node count",  # NB: N - 1
                "SF076": "BPAC area longitude grid node count",  # NB: N - 1
                "SF077": "BPAC area latitude grid node spacing",
                "SF078": "BPAC area longitude grid node spacing",
                "SF080": "Average area VTEC",
                "SF079": "Grid node present mask",  # len = (SF075 + 1) * (SF076 + 1)
                "groupVTEC": (
                    NB + "SF079+1",  # repeating group
                    {
                        "SF055": "VTEC quality",
                        "SF081": "VTEC size indicator",
                        "optSF081-0": (
                            ("SF081+2", 0),  # if SF081 = 0
                            {
                                "SF082": "Small VTEC residual",
                            },
                        ),
                        "optSF081-1": (
                            ("SF081+2", 1),  # if SF081 = 1
                            {
                                "SF083": "Large VTEC residual",
                            },
                        ),
                    },
                ),
            },
        ),
    },
    # ********************************************************************
    # EAS-DYN TODO not yet tested - no available test data source
    # ********************************************************************
    "SPARTN-1X-EAS-DYN": {
        "SF084": "Customer key ID",  # plain text
        "SF085": "Dynamic key encryption type",  # plain text
        "SF086": "Week of applicability",
        "SF085": "Payload encryption type",
        "SF087": "Dynamic key length",
        "SF088": "Dynamic key",  # variable length
    },
    # ********************************************************************
    # EAS-GRP deprecated
    # ********************************************************************
    "SPARTN-1X-EAS-GRP": {},  # no current plans to implement
}
